@@ -1,7 +1,7 @@
 /*
  * plugin-api-info.c - extra info functions for plugin API
  *
- * Copyright (C) 2003-2021 Sébastien Helleu <flashcode@flashtux.org>
+ * Copyright (C) 2003-2023 Sébastien Helleu <flashcode@flashtux.org>
  *
  * This file is part of WeeChat, the extensible chat client.
  *
@@ -32,9 +32,11 @@
 #include "../core/weechat.h"
 #include "../core/wee-config.h"
 #include "../core/wee-crypto.h"
+#include "../core/wee-hashtable.h"
 #include "../core/wee-hook.h"
 #include "../core/wee-infolist.h"
 #include "../core/wee-proxy.h"
+#include "../core/wee-secure.h"
 #include "../core/wee-string.h"
 #include "../core/wee-url.h"
 #include "../core/wee-util.h"
@@ -96,8 +98,10 @@ plugin_api_info_version_number_cb (const void *pointer, void *data,
     (void) info_name;
     (void) arguments;
 
-    snprintf (version_number, sizeof (version_number), "%d",
-              util_version_number (version_get_version ()));
+    snprintf (
+        version_number, sizeof (version_number), "%d",
+        util_version_number (
+            (arguments && arguments[0]) ? arguments : version_get_version ()));
     return strdup (version_number);
 }
 
@@ -444,6 +448,55 @@ plugin_api_info_auto_connect_cb (const void *pointer, void *data,
     return strdup (value);
 }
 
+/*
+ * Returns WeeChat info "auto_load_scripts".
+ */
+
+char *
+plugin_api_info_auto_load_scripts_cb (const void *pointer, void *data,
+                                      const char *info_name,
+                                      const char *arguments)
+{
+    char value[32];
+
+    /* make C compiler happy */
+    (void) pointer;
+    (void) data;
+    (void) info_name;
+    (void) arguments;
+
+    snprintf (value, sizeof (value), "%d", weechat_auto_load_scripts);
+    return strdup (value);
+}
+
+/*
+ * Returns WeeChat info "buffer".
+ */
+
+char *
+plugin_api_info_buffer_cb (const void *pointer, void *data,
+                           const char *info_name,
+                           const char *arguments)
+{
+    struct t_gui_buffer *ptr_buffer;
+    char value[64];
+
+    /* make C compiler happy */
+    (void) pointer;
+    (void) data;
+    (void) info_name;
+
+    if (!arguments || !arguments[0])
+        return NULL;
+
+    ptr_buffer = gui_buffer_search_by_full_name (arguments);
+    if (!ptr_buffer)
+        return NULL;
+
+    snprintf (value, sizeof (value), "0x%lx", (unsigned long)ptr_buffer);
+    return strdup (value);
+}
+
 /*
  * Returns WeeChat info "charset_terminal".
  */
@@ -762,6 +815,7 @@ plugin_api_info_nick_color_cb (const void *pointer, void *data,
 
     result = gui_nick_find_color (
         (num_items >= 1) ? items[0] : NULL,
+        -1,
         (num_items >= 2) ? items[1] : NULL);
 
     if (items)
@@ -791,6 +845,7 @@ plugin_api_info_nick_color_name_cb (const void *pointer, void *data,
 
     result = gui_nick_find_color_name (
         (num_items >= 1) ? items[0] : NULL,
+        -1,
         (num_items >= 2) ? items[1] : NULL);
 
     if (items)
@@ -799,6 +854,128 @@ plugin_api_info_nick_color_name_cb (const void *pointer, void *data,
     return result;
 }
 
+/*
+ * Returns nick color code for a nickname (case ignored using a range of chars).
+ */
+
+char *
+plugin_api_info_nick_color_ignore_case_cb (const void *pointer, void *data,
+                                           const char *info_name,
+                                           const char *arguments)
+{
+    char **items, *result, *error;
+    int num_items, case_range;
+    long number;
+
+    /* make C compiler happy */
+    (void) pointer;
+    (void) data;
+    (void) info_name;
+
+    items = string_split (arguments, ";", NULL, 0, 3, &num_items);
+
+    case_range = -1;
+    if (num_items >= 2)
+    {
+        error = NULL;
+        number = strtol (items[1], &error, 10);
+        if (error && !error[0])
+            case_range = (int)number;
+    }
+
+    result = gui_nick_find_color (
+        (num_items >= 1) ? items[0] : NULL,
+        case_range,
+        (num_items >= 3) ? items[2] : NULL);
+
+    if (items)
+        string_free_split (items);
+
+    return result;
+}
+
+/*
+ * Returns nick color name for a nickname (case ignored using a range of chars).
+ */
+
+char *
+plugin_api_info_nick_color_name_ignore_case_cb (const void *pointer, void *data,
+                                                const char *info_name,
+                                                const char *arguments)
+{
+    char **items, *result, *error;
+    int num_items, case_range;
+    long number;
+
+    /* make C compiler happy */
+    (void) pointer;
+    (void) data;
+    (void) info_name;
+
+    items = string_split (arguments, ";", NULL, 0, 3, &num_items);
+
+    case_range = -1;
+    if (num_items >= 2)
+    {
+        error = NULL;
+        number = strtol (items[1], &error, 10);
+        if (error && !error[0])
+            case_range = (int)number;
+    }
+
+    result = gui_nick_find_color_name (
+        (num_items >= 1) ? items[0] : NULL,
+        case_range,
+        (num_items >= 3) ? items[2] : NULL);
+
+    if (items)
+        string_free_split (items);
+
+    return result;
+}
+
+/*
+ * Returns uptime according to the start date and arguments.
+ */
+
+char *
+plugin_api_info_build_uptime (time_t start_time, const char *arguments)
+{
+    char value[128];
+    time_t total_seconds;
+    int days, hours, minutes, seconds;
+
+    if (!arguments || !arguments[0])
+    {
+        /* return uptime with format: "days:hh:mm:ss" */
+        util_get_time_diff (start_time, time (NULL),
+                            NULL, &days, &hours, &minutes, &seconds);
+        snprintf (value, sizeof (value), "%d:%02d:%02d:%02d",
+                  days, hours, minutes, seconds);
+        return strdup (value);
+    }
+
+    if (strcmp (arguments, "days") == 0)
+    {
+        /* return the number of days */
+        util_get_time_diff (start_time, time (NULL),
+                            NULL, &days, NULL, NULL, NULL);
+        snprintf (value, sizeof (value), "%d", days);
+        return strdup (value);
+    }
+
+    if (strcmp (arguments, "seconds") == 0)
+    {
+        /* return the number of seconds */
+        util_get_time_diff (start_time, time (NULL),
+                            &total_seconds, NULL, NULL, NULL, NULL);
+        snprintf (value, sizeof (value), "%lld", (long long)total_seconds);
+        return strdup (value);
+    }
+
+    return NULL;
+}
+
 /*
  * Returns WeeChat info "uptime".
  */
@@ -808,44 +985,32 @@ plugin_api_info_uptime_cb (const void *pointer, void *data,
                            const char *info_name,
                            const char *arguments)
 {
-    char value[128];
-    time_t total_seconds;
-    int days, hours, minutes, seconds;
-
     /* make C compiler happy */
     (void) pointer;
     (void) data;
     (void) info_name;
 
-    if (!arguments || !arguments[0])
-    {
-        /* return uptime with format: "days:hh:mm:ss" */
-        util_get_time_diff (weechat_first_start_time, time (NULL),
-                            NULL, &days, &hours, &minutes, &seconds);
-        snprintf (value, sizeof (value), "%d:%02d:%02d:%02d",
-                  days, hours, minutes, seconds);
-        return strdup (value);
-    }
+    return plugin_api_info_build_uptime (weechat_first_start_time, arguments);
+}
 
-    if (strcmp (arguments, "days") == 0)
-    {
-        /* return the number of days */
-        util_get_time_diff (weechat_first_start_time, time (NULL),
-                            NULL, &days, NULL, NULL, NULL);
-        snprintf (value, sizeof (value), "%d", days);
-        return strdup (value);
-    }
+/*
+ * Returns WeeChat info "uptime_current" (current run: from last start,
+ * upgrades are ignored).
+ */
 
-    if (strcmp (arguments, "seconds") == 0)
-    {
-        /* return the number of seconds */
-        util_get_time_diff (weechat_first_start_time, time (NULL),
-                            &total_seconds, NULL, NULL, NULL, NULL);
-        snprintf (value, sizeof (value), "%lld", (long long)total_seconds);
-        return strdup (value);
-    }
+char *
+plugin_api_info_uptime_current_cb (const void *pointer, void *data,
+                                   const char *info_name,
+                                   const char *arguments)
+{
+    /* make C compiler happy */
+    (void) pointer;
+    (void) data;
+    (void) info_name;
 
-    return NULL;
+    return plugin_api_info_build_uptime (
+        (time_t)weechat_current_start_timeval.tv_sec,
+        arguments);
 }
 
 /*
@@ -992,6 +1157,24 @@ error:
     return NULL;
 }
 
+/*
+ * Returns secured data hashtable.
+ */
+
+struct t_hashtable *
+plugin_api_info_hashtable_secured_data_cb (const void *pointer, void *data,
+                                           const char *info_name,
+                                           struct t_hashtable *hashtable)
+{
+    /* make C compiler happy */
+    (void) pointer;
+    (void) data;
+    (void) info_name;
+    (void) hashtable;
+
+    return hashtable_dup (secure_hashtable_data);
+}
+
 /*
  * Returns WeeChat infolist "bar".
  *
@@ -1035,7 +1218,7 @@ plugin_api_infolist_bar_cb (const void *pointer, void *data,
         for (ptr_bar = gui_bars; ptr_bar; ptr_bar = ptr_bar->next_bar)
         {
             if (!arguments || !arguments[0]
-                || string_match (ptr_bar->name, arguments, 0))
+                || string_match (ptr_bar->name, arguments, 1))
             {
                 if (!gui_bar_add_to_infolist (ptr_infolist, ptr_bar))
                 {
@@ -1094,7 +1277,7 @@ plugin_api_infolist_bar_item_cb (const void *pointer, void *data,
              ptr_bar_item = ptr_bar_item->next_item)
         {
             if (!arguments || !arguments[0]
-                || string_match (ptr_bar_item->name, arguments, 0))
+                || string_match (ptr_bar_item->name, arguments, 1))
             {
                 if (!gui_bar_item_add_to_infolist (ptr_infolist, ptr_bar_item))
                 {
@@ -1227,7 +1410,7 @@ plugin_api_infolist_buffer_cb (const void *pointer, void *data,
              ptr_buffer = ptr_buffer->next_buffer)
         {
             if (!arguments || !arguments[0]
-                || string_match (ptr_buffer->full_name, arguments, 0))
+                || string_match (ptr_buffer->full_name, arguments, 1))
             {
                 if (!gui_buffer_add_to_infolist (ptr_infolist, ptr_buffer))
                 {
@@ -1317,7 +1500,7 @@ plugin_api_infolist_filter_cb (const void *pointer, void *data,
          ptr_filter = ptr_filter->next_filter)
     {
         if (!arguments || !arguments[0]
-            || string_match (ptr_filter->name, arguments, 0))
+            || string_match (ptr_filter->name, arguments, 1))
         {
             if (!gui_filter_add_to_infolist (ptr_infolist, ptr_filter))
             {
@@ -1631,7 +1814,7 @@ plugin_api_infolist_plugin_cb (const void *pointer, void *data,
              ptr_plugin = ptr_plugin->next_plugin)
         {
             if (!arguments || !arguments[0]
-                || string_match (ptr_plugin->name, arguments, 0))
+                || string_match (ptr_plugin->name, arguments, 1))
             {
                 if (!plugin_add_to_infolist (ptr_infolist, ptr_plugin))
                 {
@@ -1690,7 +1873,7 @@ plugin_api_infolist_proxy_cb (const void *pointer, void *data,
              ptr_proxy = ptr_proxy->next_proxy)
         {
             if (!arguments || !arguments[0]
-                || string_match (ptr_proxy->name, arguments, 0))
+                || string_match (ptr_proxy->name, arguments, 1))
             {
                 if (!proxy_add_to_infolist (ptr_infolist, ptr_proxy))
                 {
@@ -1850,7 +2033,9 @@ plugin_api_info_init ()
                NULL, &plugin_api_info_version_cb, NULL, NULL);
     hook_info (NULL, "version_number",
                N_("WeeChat version (as number)"),
-               NULL, &plugin_api_info_version_number_cb, NULL, NULL);
+               N_("version (optional, by default the version of the running "
+                  "WeeChat is returned)"),
+               &plugin_api_info_version_number_cb, NULL, NULL);
     hook_info (NULL, "version_git",
                N_("WeeChat git version (output of command \"git describe\" "
                   "for a development version only, empty for a stable "
@@ -1910,9 +2095,18 @@ plugin_api_info_init ()
                NULL, &plugin_api_info_weechat_daemon_cb, NULL, NULL);
     hook_info (NULL, "auto_connect",
                N_("1 if automatic connection to servers is enabled, "
-                  "0 if it has been disabled by the user (option \"-a\" or "
-                  "\"--no-connect\")"),
+                  "0 if it has been disabled by the user "
+                  "(option \"-a\" or \"--no-connect\")"),
                NULL, &plugin_api_info_auto_connect_cb, NULL, NULL);
+    hook_info (NULL, "auto_load_scripts",
+               N_("1 if scripts are automatically loaded, "
+                  "0 if the auto-load has been disabled by the user "
+                  "(option \"-s\" or \"--no-script\")"),
+               NULL, &plugin_api_info_auto_load_scripts_cb, NULL, NULL);
+    hook_info (NULL, "buffer",
+               N_("buffer pointer"),
+               N_("buffer full name"),
+               &plugin_api_info_buffer_cb, NULL, NULL);
     hook_info (NULL, "charset_terminal",
                N_("terminal charset"),
                NULL, &plugin_api_info_charset_terminal_cb, NULL, NULL);
@@ -1971,11 +2165,37 @@ plugin_api_info_init ()
                   "options with nick colors and forced nick colors are "
                   "ignored)"),
                &plugin_api_info_nick_color_name_cb, NULL, NULL);
+    hook_info (NULL, "nick_color_ignore_case",
+               N_("get nick color code, ignoring case"),
+               N_("nickname;range;colors (range is a number of chars (see "
+                  "function strcasecmp_range, 0 = convert to lower case without "
+                   "using a range), colors is an optional comma-separated list "
+                  "of colors to use; background is allowed for a color with "
+                  "format text:background; if colors is present, WeeChat "
+                  "options with nick colors and forced nick colors are "
+                  "ignored)"),
+               &plugin_api_info_nick_color_ignore_case_cb, NULL, NULL);
+    hook_info (NULL, "nick_color_name_ignore_case",
+               N_("get nick color name, ignoring case"),
+               N_("nickname;range;colors (range is a number of chars (see "
+                  "function strcasecmp_range, 0 = convert to lower case without "
+                   "using a range), colors is an optional comma-separated list "
+                  "of colors to use; background is allowed for a color with "
+                  "format text:background; if colors is present, WeeChat "
+                  "options with nick colors and forced nick colors are "
+                  "ignored)"),
+               &plugin_api_info_nick_color_name_ignore_case_cb, NULL, NULL);
     hook_info (NULL, "uptime",
                N_("WeeChat uptime (format: \"days:hh:mm:ss\")"),
                N_("\"days\" (number of days) or \"seconds\" (number of "
                   "seconds) (optional)"),
                &plugin_api_info_uptime_cb, NULL, NULL);
+    hook_info (NULL, "uptime_current",
+               N_("WeeChat uptime for the current process only (upgrades with "
+                  "/upgrade command are ignored) (format: \"days:hh:mm:ss\")"),
+               N_("\"days\" (number of days) or \"seconds\" (number of "
+                  "seconds) (optional)"),
+               &plugin_api_info_uptime_current_cb, NULL, NULL);
     hook_info (NULL, "totp_generate",
                N_("generate a Time-based One-Time Password (TOTP)"),
                N_("secret (in base32), timestamp (optional, current time by "
@@ -1999,6 +2219,15 @@ plugin_api_info_init ()
            "\"y\": y coordinate (string with integer >= 0)"),
         N_("see function \"hook_focus\" in Plugin API reference"),
         &gui_focus_info_hashtable_gui_focus_info_cb, NULL, NULL);
+    /* info (hashtable) with the secured data */
+    hook_info_hashtable (
+        NULL,
+        "secured_data",
+        N_("secured data"),
+        NULL,
+        N_("secured data: names and values (be careful: the values are "
+           "sensitive data: do NOT print/log them anywhere)"),
+        &plugin_api_info_hashtable_secured_data_cb, NULL, NULL);
 
     /* WeeChat core infolist hooks */
     hook_infolist (NULL, "bar",
